import tarfile
from datetime import datetime, timedelta
import os

import numpy as np
import pytest
import requests
from tonik import Storage
from zizou.data import SDSWaveforms, DataSource
import yaml

from zizou.util import generate_test_data

def pytest_addoption(parser):
    parser.addoption(
        "--runwebservice", action="store_true", default=False, help="run webservice tests"
    )
    parser.addoption(
        "--runslow", action="store_true", default=False, help="run slow tests"
    )


def pytest_configure(config):
    config.addinivalue_line("markers", "slow: mark test as slow to run")
    config.addinivalue_line("markers", "webservice: mark tests that request data from webservices")


def pytest_collection_modifyitems(config, items):
    if not config.getoption("--runwebservice"):
        skip_webservice = pytest.mark.skip(reason="need --runwebservice option to run")
        for item in items:
             if "webservice" in item.keywords:
                item.add_marker(skip_webservice)
    if not config.getoption("--runslow"):
        skip_slow = pytest.mark.skip(reason="need --runslow option to run")
        for item in items:
            if "slow" in item.keywords:
                item.add_marker(skip_slow)
   

tstart = datetime(2023,1,1,0,0,0)
ndays = 10

@pytest.fixture(scope='package')
def setup(tmp_path_factory):
    features1D = ['rsam',
                  'dsar',
                  'central_freq',
                  'predom_freq',
                  'bandwidth',
                  'rsam_energy_prop']
    features2D = [('sonogram', 'sonofrequency'),
                  ('ssam', 'frequency'),
                  ('filterbank', 'fbfrequency')]

    savedir = tmp_path_factory.mktemp('zizou_test_tmp', numbered=True)
    sgw = Storage('volcanoes', savedir)
    s_wiz = sgw.get_substore('WIZ', '00', 'HHZ') 
    s_mdr = sgw.get_substore('MDR', '00', 'BHZ') 
    s_mavz = sgw.get_substore('MAVZ', '00', 'EHZ') 
    s_mms = sgw.get_substore('MMS', '00', 'BHZ') 
   # Generate some fake data
    for _f in features1D:
        feat = generate_test_data(tstart=tstart,
                                    feature_name=_f,
                                    ndays=ndays)
        for _s in [s_wiz, s_mdr, s_mavz, s_mms]:
            _s.save(feat)
    for _n, _f in features2D:
        feat = generate_test_data(tstart=tstart,
                                  feature_name=_n,
                                  ndays=ndays,
                                  nfreqs=8,
                                  freq_name=_f,
                                  dim=2)
        for _s in [s_wiz, s_mdr, s_mavz, s_mms]:
            _s.save(feat)

    alg = generate_test_data(tstart=tstart,
                             feature_name='autoencoder',
                             ndays=ndays,
                             nfreqs=5,
                             freq_name='cluster',
                             dim=2)
    s_mdr.save(alg)
    return savedir

@pytest.fixture(scope='module')
def setup_ac(setup):
    savedir = setup
    sg = Storage('volcanoes', rootdir=savedir,
                      starttime=tstart, 
                      endtime=tstart + timedelta(days=ndays))
    s_wiz = sg.get_substore('WIZ', '00', 'HHZ')
    s_wsrz = sg.get_substore('WSRZ', '00', 'HHZ')
    config = """ 
    autoencoder:
      layers: [2000,500,200,6]
      epochs: 5
      patience: 10
    """
    return savedir, s_wiz, s_wsrz, config


def download_and_extract_tar(url, extract_path):
    # Download the tar file
    response = requests.get(url, stream=True)
    if response.status_code == 200:
        with open('file.tar', 'wb') as f:
            f.write(response.raw.read())

        # Open the tar file
        tar = tarfile.open('file.tar')

        # Extract it
        tar.extractall(path=extract_path)
        tar.close()
    else:
        print(f"Failed to download {url}")


@pytest.fixture(scope='module')
def setup_sds():
    test_data_dir = os.path.join(os.environ['HOME'], 'zizou_test_data')
    os.makedirs(test_data_dir, exist_ok=True)
    if not os.path.exists(os.path.join(test_data_dir, 'sds_test')):
        url = "https://zenodo.org/records/13377159/files/zizou_test_data.tar.gz"
        download_and_extract_tar(url, test_data_dir)
    sds_dir = os.path.join(test_data_dir, "sds_test")
    fdsn_urls=('https://service.geonet.org.nz',
                'https://service-nrt.geonet.org.nz')
    sdsc = SDSWaveforms(sds_dir=sds_dir, fdsn_urls=fdsn_urls,
                        staxml_dir=sds_dir, fill_value=np.nan)
    ds = DataSource(clients=[sdsc])
    return sds_dir, ds 




from datetime import datetime, timedelta
import os
import shutil
import tempfile
import unittest

from tonik import Storage
import xarray as xr

from zizou import AnomalyDetectionBaseClass
from zizou.util import generate_test_data


def test_get_features(setup_ac):
    savedir, sg, _, config = setup_ac
    transform_dict = {'rsam': 'log'}
    stack_dict = {'dsar': '2D',
                  'central_freq': '1H',
                  'predom_freq': '1H',
                  'variance': '1H',
                  'bandwidth': '1H'}
    ab = AnomalyDetectionBaseClass(['rsam', 'sonogram', 'dsar'], stacks=stack_dict,
                                    transforms=transform_dict)
    featurefile = os.path.join(savedir, 'pca_features.nc')
    feats = ab.get_features(sg, featurefile)
    assert feats.shape == (10*144, 10)
    # Run again to test reading features from file
    feats1 = ab.get_features(sg, featurefile)
    xr.testing.assert_allclose(feats, feats1)



from configparser import ConfigParser
from datetime import datetime
import logging
import logging.config

import pytest

from zizou.autoencoder import AutoEncoder
from zizou import get_data
from zizou import zizou_logging


logger = logging.getLogger(__name__)


@pytest.mark.slow
def test_fit(setup_ac):
    rootdir, store1, store2, config = setup_ac
    ac = AutoEncoder(config)
    ac.fit(store1)
    assert float(ac.loss_) - 1 < 0.5


#@pytest.mark.slow
def test_transform(setup_ac):
    savedir, store1, store2, config = setup_ac
    starttime = datetime(2023,1,8,0,0,0)
    endtime = datetime(2023,1,9,0,0,0)
    store1.starttime = starttime
    store1.endtime = endtime
    ac = AutoEncoder(config)
    ed = ac.fit_transform(store1)
    store1.save(ed, mode='w')
    assert ed['autoencoder_embedding'].shape == (6, 145)
    assert ed['autoencoder_cluster'].shape == (5, 145)
    assert ed['autoencoder_loss'].shape == (145,)



@pytest.mark.slow
def test_early_stopping(setup_ac):
    savedir, store1, store2, config = setup_ac
    starttime = datetime(2023,1,8,0,0,0)
    endtime = datetime(2023,1,9,0,0,0)
    store1.starttime = starttime
    store1.endtime = endtime
    ac = AutoEncoder(config)
    ed = ac.fit_transform(store1)
    assert ed['autoencoder_embedding'].shape == (6, 145)
    assert ed['autoencoder_cluster'].shape == (5, 145)
    assert ed['autoencoder_loss'].shape == (145,)


def test_exceptions(setup_ac):
    rootdir, store1, store2, config = setup_ac
    ac = AutoEncoder(config)
    with pytest.raises(AssertionError):
        ac.transform(store2)


 


import datetime
import inspect
from io import StringIO
import json
import os
import shutil
import tempfile
import unittest
from unittest.mock import patch, Mock

import pytest
import boto3
from moto import mock_s3
import numpy as np
import pandas as pd
from obspy import UTCDateTime, read_inventory, Trace
import xarray as xr
from zizou.util import generate_test_data, test_signal


from zizou.data import (WaveformBaseclass,
                        DataSource,
                        SDSWaveforms,
                        S3Waveforms,
                        MockSDSWaveforms,
                        FDSNWaveforms,
                        VolcanoMetadata,
                        PostProcess,
                        PostProcessException)
from zizou import get_data

def get_max(trace):
    value = np.nanmax(trace.data)
    _min = np.nanmin(trace.data)
    if abs(_min) > abs(value):
        value = _min
    return value


def test_exceptions():
    """
    Test that the right exceptions are raised.
    """
    with pytest.raises(ValueError):
        ds = DataSource(clients=())

    with pytest.raises(ValueError):
        ds = DataSource(clients=('blub', 'blab'))

    with pytest.raises(ValueError):
        ds = FDSNWaveforms(url='/foo/bar')


@pytest.mark.webservice
def test_get_waveforms(setup_sds):
    """
    Test that the waveforms returned from SDS and FDSNws are
    consistent.
    """
    sds_dir, _ = setup_sds
    fdsn_urls=('https://service.geonet.org.nz',
                'https://service-nrt.geonet.org.nz')
    sdsc = SDSWaveforms(sds_dir=sds_dir, fdsn_urls=fdsn_urls,
                        staxml_dir=sds_dir, fill_value=np.nan)
    fdsn_clients = []
    for url in fdsn_urls: 
        fdsn_clients.append(FDSNWaveforms(url=url, fill_value=np.nan))

    ds1 = DataSource(clients=fdsn_clients)
    ds2 = DataSource(clients=[sdsc])
    # dates are inclusive, and complete days
    starttime = UTCDateTime(year=2013, julday=240)
    endtime = starttime + 86400.
    for tr1 in ds1.get_waveforms('NZ', 'WIZ', '10', 'HHZ', starttime,
                                    endtime):
        for tr2 in ds2.get_waveforms('NZ', 'WIZ', '10', 'HHZ', starttime,
                                        endtime):
            assert get_max(tr1) == get_max(tr2)
            assert tr1.stats.npts == tr2.stats.npts


@pytest.mark.webservice
def test_with_gaps(setup_sds):
    """
    Test behaviour of traces with gaps.
    """
    sds_dir, _ = setup_sds
    fdsn_urls=('https://service.geonet.org.nz',
                'https://service-nrt.geonet.org.nz'),
    sdsc = SDSWaveforms(sds_dir=sds_dir, fdsn_urls=fdsn_urls,
                        staxml_dir=sds_dir, fill_value=np.nan)
    ds1 = DataSource(clients=[sdsc])
    starttime = UTCDateTime(2010, 11, 26, 0, 0, 0)
    endtime = starttime + 86400.
    stream = 'NZ.WIZ.10.HHZ'
    net, site, loc, comp = stream.split('.')
    gen = ds1.get_waveforms(net, site, loc, comp, starttime, endtime)
    t1 = next(gen)
    assert np.any(np.isnan(t1.data)) == True
    idx = np.where(np.isnan(t1.data))
    assert idx[0].size == 431238

    starttime2 = UTCDateTime(2010, 11, 26, 0, 0, 0)
    endtime2 = starttime2 + 86400.
    gen2 = ds1.get_waveforms(net, site, loc, comp, starttime, endtime)
    t2 = next(gen2)

    # Check that we can take a trace apart and put it
    # together again
    t3 = t2.copy()
    mdata = np.ma.masked_invalid(t3.data)
    t3.data = mdata
    st = t3.split()
    st.merge(fill_value=np.nan)
    st.trim(starttime2, endtime2, pad=True, fill_value=np.nan,
            nearest_sample=False)
    idx = np.where(st[0].data != t2.data)
    # data 'differs' only at NaNs
    assert np.alltrue(np.isnan(st[0].data[idx])) == True


@pytest.mark.webservice
def test_start_endtime(setup_sds):
    """
    Test that start and end times are what we expect.
    """
    sds_dir, _ = setup_sds
    fdsn_urls=('https://service.geonet.org.nz',
                'https://service-nrt.geonet.org.nz'),
    sdsc = SDSWaveforms(sds_dir=sds_dir, fdsn_urls=fdsn_urls,
                        staxml_dir=sds_dir, fill_value=np.nan)
    ds = DataSource(clients=[sdsc])
    starttime = UTCDateTime(year=2013, julday=240)
    endtime = starttime + 86400.
    stream = 'NZ.WIZ.10.HHZ'
    net, site, loc, comp = stream.split('.')
    gen = ds.get_waveforms(net, site, loc, comp, starttime, endtime)
    tr = next(gen)
    assert tr.stats.starttime >= starttime
    assert tr.stats.endtime <= endtime
    fdsnc = FDSNWaveforms(url='https://service.geonet.org.nz')
    ds1 = DataSource(clients=[fdsnc])
    gen1 = ds1.get_waveforms(net, site, loc, comp, starttime, endtime)
    tr1 = next(gen1)
    assert tr1.stats.starttime >= starttime
    assert tr1.stats.endtime <= endtime
 

@pytest.mark.slow
def test_missing_station_metadata_sds(setup_sds):
    """
    At NTVZ, station metadata stops at
    2019-04-17T01:10:00.000000Z and restarts at
    2019-04-19T01:30:00.000000Z, but waveform data
    still exist throughout.

    Full get_waveform test across three days, which checks that traces
    are available for the time periods where there are both station
    and waveform data, np.nan otherwise.
    """
    sds_dir, _ = setup_sds
    fdsn_urls=('https://service.geonet.org.nz',
                'https://service-nrt.geonet.org.nz')
    sdsc = SDSWaveforms(sds_dir=sds_dir, fdsn_urls=fdsn_urls,
                        staxml_dir=sds_dir, fill_value=np.nan)
    ds = DataSource(clients=[sdsc], chunk_size=86400.)
    starttime = UTCDateTime(2019, 4, 17)
    endtime = UTCDateTime(2019, 4, 20)
    net, site, loc, comp = ('NZ', 'NTVZ', '10', 'HHZ')
    gen = ds.get_waveforms(net, site, loc, comp, starttime, endtime)
    # number of non-NaN entries in each trace
    # 2019-04-17
    tr1 = next(gen)
    assert np.count_nonzero(~np.isnan(tr1.data)) == 371902
    # 2019-04-18
    tr2 = next(gen)
    assert np.count_nonzero(~np.isnan(tr2.data)) == 0
    # 2019-04-19
    tr3 = next(gen)
    assert np.count_nonzero(~np.isnan(tr3.data)) == 8100000
        

@pytest.mark.slow
@mock_s3
def test_aws_backend(setup_sds):
    """
    Test retrieving waveform data from AWS S3.
    """
    sds_dir, _ = setup_sds
    ### Setup virtual S3
    bucket_name = 'virtual-geonet-open-data'
    client = boto3.client('s3', region_name='us-east-1',
                            aws_access_key_id='fake_access_key',
                            aws_secret_access_key='fake_secret_key')
    client.create_bucket(Bucket=bucket_name)
    starttime1 = UTCDateTime(2019, 4, 17, 11, 0, 17)
    endtime1 = UTCDateTime(2019, 4, 17, 23, 59, 59, 999999)
    starttime2 = UTCDateTime(2019, 4, 18, 0, 0, 0)
    endtime2 = UTCDateTime(2019, 4, 18, 3, 4, 5)
    key1 = "waveforms/miniseed/2019/2019.107/MAVZ.NZ/2019.107.MAVZ.10-HHZ.NZ.D"
    key2 = "waveforms/miniseed/2019/2019.108/MAVZ.NZ/2019.108.MAVZ.10-HHZ.NZ.D"
    test_data1 = test_signal(starttime=starttime1,
                            nsec=endtime1-starttime1,
                            station='MAVZ',
                            location='10') 
    test_data2 = test_signal(starttime=starttime2,
                            nsec=endtime2-starttime2,
                            station='MAVZ',
                            location='10') 
    filename = tempfile.mktemp()
    test_data1.write(filename, format='MSEED')
    client.upload_file(Filename=filename, Bucket=bucket_name, Key=key1)
    filename = tempfile.mktemp()
    test_data2.write(filename, format='MSEED')
    client.upload_file(Filename=filename, Bucket=bucket_name, Key=key2)
    fdsn_urls=('https://service.geonet.org.nz',
                'https://service-nrt.geonet.org.nz')
    awsc = S3Waveforms(s3bucket=bucket_name, fdsn_urls=fdsn_urls,
                       staxml_dir=sds_dir, fill_value=np.nan)

    ds = DataSource(clients=[awsc])
    net, site, loc, comp = ('NZ', 'MAVZ', '10', 'HHZ')
    gen = ds.get_waveforms(net, site, loc, comp, starttime1,
                            endtime2)
    tr1 = next(gen)
    test_data1.trim(starttime1, starttime1 + datetime.timedelta(days=1))
    test_inv = read_inventory(os.path.join(sds_dir, 'MAVZ.xml'))
    test_data1.attach_response(test_inv)
    test_data1.remove_sensitivity()
    assert test_data1.data.max() == np.nanmax(tr1.data)
    with pytest.raises(StopIteration):
        tr2 = next(gen)
    assert tr1.stats.starttime == starttime1
    assert tr1.stats.endtime == endtime2
    gen1 = ds.get_waveforms(net, 'WIZ', loc, comp, starttime1,
                            endtime2)
    tr3 = next(gen1)
    # Check that the process returned a dummy trace
    assert len(tr3) == 0


def test_mock_client():
    """
    Test that the mock client is working.
    """
    sds_dir = tempfile.mkdtemp()
    msc = MockSDSWaveforms(sds_dir=sds_dir)
    ds = DataSource(clients=[msc])
    net, site, loc, comp = ('NZ', 'MAVZ', '10', 'HHZ')
    starttime = UTCDateTime(2024, 4, 11, 3, 0, 0)
    endtime = UTCDateTime(2024, 4, 11, 4, 0, 0)
    gen = ds.get_waveforms(net, site, loc, comp, starttime, endtime)
    tr = next(gen)
    assert tr.stats.starttime == starttime
    assert tr.stats.endtime == endtime


def test_caching():
    tr_orig = test_signal(nsec=4320*600., sampling_rate=1/600.,
                          starttime=UTCDateTime(2020, 7, 18, 8, 22))
    class FakeClient(WaveformBaseclass):
        def __init__(self, trace):
            self.trace = trace

        def get_waveforms(self, net, site, loc, comp, starttime, endtime):
            return self.trace.slice(starttime, endtime)
    
    cache_dir = tempfile.mkdtemp()

    ds = DataSource(clients=[FakeClient(tr_orig)], cache_dir=cache_dir)

    starttime = UTCDateTime(2020, 7, 18, 8, 22)
    endtime = UTCDateTime(2020, 8, 17, 8, 12)
    tr = next(ds.get_waveforms('NZ', 'BLUB', '', 'HHZ',
                                starttime, endtime, cache=True))

    ds_cache = DataSource(clients=[FakeClient(tr_orig)], cache_dir=cache_dir)
    tr_cache = next(ds_cache.get_waveforms('NZ', 'BLUB', '', 'HHZ',
                                           starttime, endtime, cache=True))  
    assert tr_cache.stats.cached == True
    np.alltrue(tr_orig.data == tr_cache.data)

    ds_chunked = DataSource(clients=[FakeClient(tr_orig)], cache_dir=cache_dir,
                            chunk_size=86400.)
    tr_chunked = next(ds_chunked.get_waveforms('NZ', 'BLUB', '', 'HHZ',
                                               starttime, endtime, cache=True))
    assert tr_chunked.stats.cached == True
    np.alltrue(tr_orig.slice(starttime, starttime + 86400.).data == tr_chunked.data)

"""
Tests various permutations of station metadata availability e.g. at
MAVZ:
[Channel 'HHZ', Location '10',
    Time range: 2012-05-22T02:00:00.000000Z - 2013-02-26T00:00:00.000000Z,
    Channel 'HHZ', Location '10',
    Time range: 2013-03-22T00:00:00.000000Z - 2013-04-09T22:15:00.000000Z,
    Channel 'HHZ', Location '10',
    Time range: 2013-04-10T00:15:00.000000Z - 2015-01-09T00:00:00.000000Z,
    Channel 'HHZ', Location '10',
    Time range: 2015-01-09T00:00:01.000000Z - 2017-04-20T22:00:00.000000Z,
    Channel 'HHZ', Location '10',
    Time range: 2017-04-20T22:00:02.000000Z - 9999-01-01T00:00:00.000000Z,
]
"""

def test_function_list():
    """
    Tests whether there is a matching number of tests and
    post-processing functions
    """
    pp = PostProcess()
    check_names = [check_n for (check_n, check) in pp.checklist]
    func_names = [func_n for (func_n, func) in pp.pp_functions]
    # Check correct number of checks and output function
    assert len(check_names) == len(func_names)
    # Check checks and output functions match
    for c, f in zip(check_names, func_names):
        assert c.replace('check', 'output') == f


def test_case_single_trace_in_stream(setup_sds):
    sds_dir, ds = setup_sds
    net, site, loc, comp = ('NZ', 'MAVZ', '10', 'HHZ')
    startdate = UTCDateTime(2013, 2, 25)
    enddate = UTCDateTime(2013, 2, 26)
    st = ds.clients[0].client.get_waveforms(net, site, loc, comp, startdate,
                                            enddate, dtype='float64')
    inv = read_inventory(os.path.join(sds_dir, 'MAVZ.xml'),
                            format='STATIONXML')
    inv_dt = inv.select(location=loc, channel=comp,
                        starttime=startdate, endtime=enddate)
    pp = PostProcess(st=st, inv=inv, inv_dt=inv_dt, loc=loc,
                        comp=comp, fill_value=np.nan, startdate=startdate,
                        enddate=enddate)
    pp.run_checks()
    # Check correct test case:
    for (a, b) in zip(pp.checklist, pp.res):
        if b:
            assert a[0] == "check_case_no_station_issues"
    for (a, b) in zip(pp.pp_functions, pp.res):
        if b:
            assert a[0] == "output_case_no_station_issues"
    # Check output
    test_st = st.copy()
    test_st.attach_response(inv_dt)
    test_st.remove_sensitivity()
    test_st.trim(
        starttime=startdate, endtime=enddate,
        nearest_sample=False, pad=True, fill_value=np.nan
    )
    tr = pp.run_post_processing()
    np.testing.assert_array_almost_equal(tr.data, test_st[0].data)


def test_multiple_traces_in_stream(setup_sds):
    sds_dir, ds = setup_sds
    net, site, loc, comp = ('NZ', 'MAVZ', '10', 'HHZ')
    startdate = UTCDateTime(2012, 5, 23)
    enddate = UTCDateTime(2012, 5, 24)
    st = ds.clients[0].client.get_waveforms(net, site, loc, comp, startdate,
                                            enddate, dtype='float64')
    inv = read_inventory(os.path.join(sds_dir, 'MAVZ.xml'),
                         format='STATIONXML')
    inv_dt = inv.select(location=loc, channel=comp,
                        starttime=startdate, endtime=enddate)
    pp = PostProcess(st=st, inv=inv, inv_dt=inv_dt, loc=loc,
                        comp=comp, fill_value=np.nan, startdate=startdate,
                        enddate=enddate)
    pp.run_checks()

    for (a, b) in zip(pp.checklist, pp.res):
        if b:
            assert a[0] == "check_case_no_station_issues"
    for (a, b) in zip(pp.pp_functions, pp.res):
        if b:
            assert a[0] == "output_case_no_station_issues"
    tr = pp.run_post_processing()
    assert int(tr.stats.npts) == int(8640000)

def test_case_stream_starttime_equals_channel_endtime(setup_sds):
    sds_dir, ds = setup_sds
    net, site, loc, comp = ('NZ', 'MAVZ', '10', 'HHZ')
    startdate = UTCDateTime(2013, 2, 26)
    enddate = UTCDateTime(2013, 2, 27)
    st = ds.clients[0].client.get_waveforms(net, site, loc, comp, startdate,
                                                    enddate, dtype='float64')
    inv = read_inventory(os.path.join(sds_dir, 'MAVZ.xml'),
                            format='STATIONXML')
    inv_dt = inv.select(location=loc, channel=comp,
                        starttime=startdate, endtime=enddate)
    pp = PostProcess(st=st, inv=inv, inv_dt=inv_dt, loc=loc,
                        comp=comp, fill_value=np.nan, startdate=startdate,
                        enddate=enddate)
    pp.run_checks()

    # Check correct test case:
    for (a, b) in zip(pp.checklist, pp.res):
        if b:
            assert a[0] == "check_case_no_data"
    for (a, b) in zip(pp.pp_functions, pp.res):
        if b:
            assert a[0] == "output_case_no_data"
    with pytest.raises(PostProcessException):
        pp.run_post_processing()

def test_case_stream_endtime_equals_channel_starttime(setup_sds):
    sds_dir, ds = setup_sds
    net, site, loc, comp = ('NZ', 'MAVZ', '10', 'HHZ')
    startdate = UTCDateTime(2013, 3, 21)
    enddate = UTCDateTime(2013, 3, 22)
    st = ds.clients[0].client.get_waveforms(net, site, loc, comp, startdate,
                                                    enddate, dtype='float64')
    inv = read_inventory(os.path.join(sds_dir, 'MAVZ.xml'),
                            format='STATIONXML')
    inv_dt = inv.select(location=loc, channel=comp,
                        starttime=startdate, endtime=enddate)
    pp = PostProcess(st=st, inv=inv, inv_dt=inv_dt, loc=loc,
                        comp=comp, fill_value=np.nan, startdate=startdate,
                        enddate=enddate)
    pp.run_checks()

    # Check correct test case:
    for (a, b) in zip(pp.checklist, pp.res):
        if b:
            assert a[0] == "check_case_no_data"
    for (a, b) in zip(pp.pp_functions, pp.res):
        if b:
            assert a[0] == "output_case_no_data"
    with pytest.raises(PostProcessException):
        pp.run_post_processing()

def test_case_multiple_channel_periods(setup_sds):
    sds_dir, ds = setup_sds
    net, site, loc, comp = ('NZ', 'MAVZ', '10', 'HHZ')
    startdate = UTCDateTime(2015, 1, 9)
    enddate = UTCDateTime(2015, 1, 10)
    st = ds.clients[0].client.get_waveforms(net, site, loc, comp, startdate,
                                                    enddate, dtype='float64')
    inv = read_inventory(os.path.join(sds_dir, 'MAVZ.xml'),
                            format='STATIONXML')
    inv_dt = inv.select(location=loc, channel=comp,
                        starttime=startdate, endtime=enddate)
    pp = PostProcess(st=st, inv=inv, inv_dt=inv_dt, loc=loc,
                        comp=comp, fill_value=np.nan, startdate=startdate,
                        enddate=enddate)
    pp.run_checks()

    # Check correct test case:
    for (a, b) in zip(pp.checklist, pp.res):
        if b:
            assert a[0] == "check_case_multiple_channel_periods"
    for (a, b) in zip(pp.pp_functions, pp.res):
        if b:
            assert a[0] == "output_case_multiple_channel_periods"

def test_case_incomplete_station_metadata(setup_sds):
    sds_dir, ds = setup_sds
    startdate = UTCDateTime(2013, 4, 9)
    enddate = UTCDateTime(2013, 4, 10)
    net, site, loc, comp = ('NZ', 'MAVZ', '10', 'HHZ')
    inv = read_inventory(os.path.join(sds_dir, 'MAVZ.xml'),
                            format='STATIONXML')
    inv_dt = inv.select(location=loc, channel=comp,
                        starttime=startdate, endtime=enddate)
    st = ds.clients[0].client.get_waveforms(net, site, loc, comp, startdate,
                                                    enddate, dtype='float64')
    pp = PostProcess(st=st, inv=inv, inv_dt=inv_dt, loc=loc,
                        comp=comp, fill_value=np.nan, startdate=startdate,
                        enddate=enddate)
    pp.run_checks()

    # Check correct test case:
    for (a, b) in zip(pp.checklist, pp.res):
        if b:
            assert a[0] == "check_case_incomplete_station_metadata"
    for (a, b) in zip(pp.pp_functions, pp.res):
        if b:
            assert a[0] == "output_case_incomplete_station_metadata"
    test_st = st.copy()
    test_st.trim(startdate, inv_dt[0][0][0].end_date,
                    nearest_sample=False)
    test_st.attach_response(inv_dt)
    test_st.remove_sensitivity()
    test_st.trim(startdate, enddate, nearest_sample=False,
                    pad=True, fill_value=np.nan
                    )
    tr = pp.run_post_processing()

    np.testing.assert_array_almost_equal(tr.data, test_st[0].data)


class MetadataTestCase(unittest.TestCase):

    def setUp(self):
        self.data_dir = os.path.join(os.path.dirname(os.path.abspath(
            inspect.getfile(inspect.currentframe()))), "data")
        self.volcano = 'Ruapehu'
        self.eruptions = [
            datetime.date(2006, 10, 4),
            datetime.date(2007, 9, 25)
        ]
        self.st_names = [
            "NZ.MAVZ.10.HHZ",
            "NZ.WHVZ.10.HHZ",
            "NZ.TRVZ.10.HHZ",
            "NZ.FWVZ.10.HHZ",
            "NZ.COVZ.10.HHZ",
        ]

    def test_get_streams(self):
        """
        Test that the streams returned from VolcanoMetadata class are
        consistent.
        """
        file = get_data('package_data/config.yml')
        vm = VolcanoMetadata(file)
        streams = vm.get_available_streams(self.volcano)
        for stream in streams:
            self.assertTrue(stream in self.st_names)

    def test_get_eruptions(self):
        """
        Test that the eruptions dates returned from VolcanoMetadata class
        are consistent.
        """
        file = get_data('package_data/config.yml')
        vm = VolcanoMetadata(file)
        er = vm.get_eruption_dates(self.volcano)
        for e in er:
            self.assertTrue(e in self.eruptions)

    def test_find_station_coordinates(self):
        # Sample data for testing
        test_data = """ 
        metadata:
          volcano: 
          - name: Test Volcano
            network:
            - net_code: TZ
              stations:
              - sta_code: TST1
                latitude: -37.000
                longitude: 177.000
              - sta_code: TST2
                latitude: -38.000
                longitude: 178.000
        """
        vm = VolcanoMetadata(test_data)

        # Test case for an existing station
        assert vm.get_site_information("TST1")['latitude'] == -37.0
        assert vm.get_site_information("TST1")['longitude'] == 177.0

        # Test case for a non-existing station
        assert vm.get_site_information("NONEXISTENT") is None


if __name__ == '__main__':
    unittest.main()



from datetime import (date,
                      datetime,
                      timedelta,
                      timezone)
import io
import json
import logging
import os
import re
import inspect

import boto3
from botocore import UNSIGNED
from botocore.config import Config
from botocore.client import ClientError
import numpy as np
from obspy.clients.filesystem.sds import Client as SDS_Client
from obspy.clients.fdsn import Client as FDSN_Client
from obspy.clients.fdsn.header import FDSNException

from obspy import (read_inventory,
                   Inventory,
                   Trace,
                   Stream,
                   UTCDateTime,
                   read)
import pandas as pd
import requests
import xarray as xr
import yaml

from zizou.util import test_signal


logger = logging.getLogger(__name__)


class PostProcessException(Exception):
    pass

class PostProcess:
    """
    Class for checking waveforms and station metadata for valid values,
    filling waveform data gaps and removing instrument sensitivity
    """
    def __init__(self, st=Stream(), inv=Inventory(), inv_dt=Inventory(),
                 startdate=None, enddate=None, loc='*', comp='*Z',
                 fill_value=np.nan):
        self.st = st
        self.inv = inv
        self.inv_dt = inv_dt
        self.startdate = startdate
        self.enddate = enddate
        self.loc = loc
        self.comp = comp
        self.fill_value = fill_value

        # Get check registry
        self.checklist = [
            x for x in inspect.getmembers(self)
            if inspect.ismethod(x[1]) and x[0].startswith("check")
        ]
        # Get matching post-processing functions
        self.pp_functions = []
        for check in self.checklist:
            self.pp_functions.append(
                [x for x in inspect.getmembers(self)
                 if x[0] == check[0].replace("check", "output")
                 ][0]
            )

        # Run checks
        self.res = None

    def run_post_processing(self):
        self.res = self.run_checks()
        func = [a[1] for (a, b) in zip(self.pp_functions, self.res) if b][0]
        trace = func()
        return trace

    def run_checks(self):
        self.res = [check[1]() for check in self.checklist]
        if sum(self.res) == 0:
            raise NotImplementedError(
                'No post processing not set up for this case:\n'
                '{}'.format(self.st)
            )
        elif sum(self.res) > 1:
            msg = 'More than one check is true: {}. Processing workflow ' \
                  'is unclear.'.format(', '.join(
                [a[0] for (a, b) in zip(self.checklist, self.res) if b])
            )
            raise PostProcessException(msg)
        else:
            return self.res

    def check_case_no_data(self):
        """
        Checks if either no stream, or no station metadata exist.
        """
        if len(self.st) < 1 or len(self.inv_dt) < 1:
            return True
        elif len(self.inv_dt[0][0]) == 1 and \
                self.inv_dt[0][0][0].start_date == self.enddate:
            return True
        elif len(self.inv_dt[0][0]) == 1 and \
                self.inv_dt[0][0][0].end_date == self.startdate:
            return True
        else:
            return False

    def output_case_no_data(self):
        raise  PostProcessException

    def check_case_multiple_channel_periods(self):
        """
        Checks if there are multiple channel operation periods in a given
        time period. If so, different response removal is required.
        """
        if len(self.inv_dt) < 1 or len(self.st) < 1:
            return False
        elif len(self.inv_dt[0][0]) > 1:
            return True
        else:
            return False

    def output_case_multiple_channel_periods(self):
        """
        Stitches two response periods together to make a single merged
        trace.
        """
        if len(self.inv_dt[0][0]) > 2:
            raise NotImplementedError(
                'Current post processing function only valid for a '
                'maximum of two response periods.'
            )
        st1 = self.st.copy()
        st1.merge(fill_value=self.fill_value)
        st2 = st1.copy()
        st1.trim(starttime=self.startdate, endtime=self.inv_dt[0][0][0].end_date,
                 nearest_sample=False)
        st2.trim(starttime=self.inv_dt[0][0][1].start_date, endtime=self.enddate,
                 nearest_sample=False)
        st1.attach_response(self.inv_dt)
        st1.remove_sensitivity()
        st2.attach_response(self.inv_dt)
        st2.remove_sensitivity()
        st1 += st2
        st1.merge(fill_value=self.fill_value)
        st1.trim(
            self.startdate, self.enddate, nearest_sample=False,
            pad=True, fill_value=self.fill_value
        )
        return st1[0]

    def check_case_incomplete_station_metadata(self):
        """
        The case where a a channel period start/stops during the
        selected time window. Some parts of the trace in the stream
        might not correspond to a valid period according to station
        metadata.
        """
        if len(self.inv_dt) < 1:
            return False
        elif len(self.inv_dt[0][0]) != 1 or len(self.st) < 1:
            return False
        elif ((self.startdate < self.inv_dt[0][0][0].start_date < self.enddate) or
                (self.startdate < self.inv_dt[0][0][0].end_date < self.enddate)):
            return True
        else:
            return False

    def output_case_incomplete_station_metadata(self):
        if self.inv_dt[0][0][0].start_date > self.startdate:
            self.st.trim(self.inv_dt[0][0][0].start_date, self.enddate,
                         nearest_sample=False)
            self.st.attach_response(self.inv_dt)
            self.st.remove_sensitivity()
            self.st.trim(
                self.startdate, self.enddate, nearest_sample=False,
                pad=True, fill_value=self.fill_value
            )
            return self.st[0]
        elif self.inv_dt[0][0][0].end_date < self.enddate:
            self.st.trim(self.startdate, self.inv_dt[0][0][0].end_date,
                         nearest_sample=False)
            self.st.attach_response(self.inv_dt)
            self.st.remove_sensitivity()
            self.st.trim(
                self.startdate, self.enddate, nearest_sample=False,
                pad=True, fill_value=self.fill_value
            )
            return self.st[0]
        else:
            raise PostProcessException('Post-procesing function not equipped to'
                            'handle:\n'
                            '{}'.format(self.st)
                            )

    def check_case_no_station_issues(self):
        if len(self.inv_dt) < 1:
            return False
        elif len(self.inv_dt[0][0]) != 1 or len(self.st) < 1:
            return False
        elif len(self.inv_dt[0][0]) == 1 and \
                self.inv_dt[0][0][0].end_date == self.startdate:
            return False
        elif len(self.inv_dt[0][0]) == 1 and \
                self.inv_dt[0][0][0].start_date == self.enddate:
            return False
        elif (self.startdate < self.inv_dt[0][0][0].start_date < self.enddate) \
                or (self.startdate < self.inv_dt[0][0][0].end_date < self.enddate):
            return False
        else:
            return True

    def output_case_no_station_issues(self):
        self.st.merge(fill_value=self.fill_value)
        self.st.trim(
            starttime=self.startdate, endtime=self.enddate,
            nearest_sample=False, pad=True, fill_value=self.fill_value
        )
        self.st.attach_response(self.inv_dt)
        self.st.remove_sensitivity()
        return self.st[0]


def get_instrument_response(net, site, loc,
                            staxml_dir='./instrument_response',
                            fdsn_urls=('https://service.geonet.org.nz')):
    """
    Retrieve instrument response file in STATIONXML format.
    """
    fn = '{:s}.xml'.format(site)
    fn = os.path.join(staxml_dir, fn)
    try:
        inv = read_inventory(fn, format='STATIONXML')
    except FileNotFoundError:
        try:
            os.makedirs(staxml_dir)
        except FileExistsError:
            pass
        for url in fdsn_urls:
            try:
                client = FDSN_Client(base_url=url)
                inv = client.get_stations(network=net,
                                          location=loc,
                                          station=site,
                                          channel='*',
                                          level='response')
            except Exception as e:
                logger.info(e)
            else:
                break
        inv.write(fn, format='STATIONXML')
    return inv


class WaveformBaseclass:
    pass

class SDSWaveforms(WaveformBaseclass):

    def __init__(self, sds_dir, fdsn_urls, staxml_dir,
                 fill_value=np.nan):
        """
        Get seismic waveforms from a local SDS archive.
        """
        self.client = SDS_Client(sds_dir)
        self.fdsn_urls = fdsn_urls
        self.staxml_dir = staxml_dir
        self.fill_value = fill_value

    def get_waveforms(self, net, site, loc, comp, startdate, enddate):
        """
        :param fill_value: Default value used to fill gaps and pad traces to
                           cover the whole time period defined by `startdate`
                           and `enddate`. If 'interpolate' is chosen, the
                           traces will not be padded.
        :type fill_value: int, float, 'interpolate', np.nan, or None
        """
        inv = get_instrument_response(net, site, loc,
                                      staxml_dir=self.staxml_dir,
                                      fdsn_urls=self.fdsn_urls)

        st = self.client.get_waveforms(net, site, loc, comp, startdate,
                                       enddate, dtype='float64')

        inv_dt = inv.select(location=loc, channel=comp,
                            starttime=startdate,
                            endtime=enddate - 1)

        # Post-process to remove sensitivity and account for gaps
        pp = PostProcess(st=st, inv=inv, inv_dt=inv_dt,
                         startdate=startdate, enddate=enddate,
                         loc=loc, comp=comp, fill_value=self.fill_value)
        tr = pp.run_post_processing()
        return tr


class MockSDSWaveforms(WaveformBaseclass):
    """
    Mock SDSWaveforms class for testing by creating
    synthetic data for the requested streams.
    """

    def __init__(self, sds_dir):
        """
        Get seismic waveforms from a local SDS archive.
        """
        os.makedirs(sds_dir, exist_ok=True)
        self.client = SDS_Client(sds_dir)

    def save2sds(self, trace: Trace, rootdir: str):
        """
        Save a trace to a SDS directory structure.

        Parameters
        ----------
        trace : `obspy.Trace` 
            Seismic/acoustic trace to be written to disk. 
        rootdir : str 
            Root directory for the SDS directory structure. 
        """
        sds_fmtstr = os.path.join(
            "{year}", "{network}", "{station}", "{channel}.{sds_type}",
            "{network}.{station}.{location}.{channel}.{sds_type}.{year}.{doy:03d}")

        fullpath = sds_fmtstr.format(year=trace.stats.starttime.year,
                                    doy=trace.stats.starttime.julday,
                                    sds_type="D", **trace.stats)
        fullpath = os.path.join(rootdir, fullpath)
        dirname, filename = os.path.split(fullpath)
        os.makedirs(dirname, exist_ok=True)
        print("writing", fullpath)
        trace.write(fullpath, format="mseed")


    def generate_dataset(self, rootdir: str, net: str, site: str,
                         loc: str, comp: str, start: str, end: str) -> str:
        """ 
        Generate a test dataset for the integration tests.

        Parameters
        ----------
        rootdir : str
            Parent directory for the test dataset.
        start : str
            Start time for the test dataset in ISO 8601 format.
        end : str
            End time for the test dataset in ISO 8601 format.
        """ 
        tstart = UTCDateTime(start)
        # Always generate whole day files
        _tstart = UTCDateTime(year=tstart.year, julday=tstart.julday)
        tend = UTCDateTime(end)
        tr = test_signal(starttime=_tstart,
                        sampling_rate=10,
                        nsec=86400, gaps=True,
                        network=net, station=site,  
                        location=loc, channel=comp)
        while _tstart < tend:
            tr.stats.starttime = _tstart
            self.save2sds(tr, rootdir)
            _tstart += 86400


    def get_waveforms(self, net, site, loc, comp, startdate, enddate):
        """
        Generate synthetic daily files that start at midnight on the
        startdate and end at midnight on the enddate. The return the requested
        timespan from these files.
        """
        self.generate_dataset(self.client.sds_root, net, site, loc, comp,
                              startdate, enddate)
        st = self.client.get_waveforms(net, site, loc, comp, startdate,
                                       enddate, dtype='float64')
        return st[0]


class FDSNWaveforms(WaveformBaseclass):

    def __init__(self, url, debug=False, fill_value=np.nan):
        """
        Get seismic waveforms from FDSN web service.
        """
        self.client = FDSN_Client(base_url=url, debug=debug)
        self.fill_value = fill_value

    def get_waveforms(self, net, site, loc, comp, startdate, enddate):
        st = self.client.get_waveforms(net, site, loc, comp, startdate,
                                       enddate, attach_response=True)
        # Prepare seismic time series
        st.remove_sensitivity()
        # in case stream has more than one trace
        st.merge(fill_value=self.fill_value)
        if self.fill_value != 'interpolate':
            st.trim(startdate, enddate, pad=True, fill_value=self.fill_value,
                    nearest_sample=False)
        return st[0]


class S3Waveforms(WaveformBaseclass):

    def __init__(self, s3bucket, fdsn_urls, staxml_dir, debug=False,
                 fill_value=np.nan):
        config = Config(signature_version=UNSIGNED)
        self.client = boto3.client('s3', config=config)
        self.s3_bucket = s3bucket 
        self.fdsn_urls = fdsn_urls
        self.staxml_dir = staxml_dir
        self.debug = debug
        self.fill_value = fill_value


    def get_waveforms(self, net, site, loc, comp, start, end):
        """
        Get seismic waveforms from GeoNet's open data archive on AWS S3.
        """
        PATH_FORMAT = "waveforms/miniseed/{year}/{year}.{julday:03d}/{station}."
        PATH_FORMAT += "{network}/{year}.{julday:03d}.{station}."
        PATH_FORMAT += "{location}-{channel}.{network}.D"
        t_start = start
        t_end = UTCDateTime(year=t_start.year,
                            julday=t_start.julday + 1,
                            hour=0, minute=0,
                            second=0, microsecond=0)
        t_end = min(t_end, end)
        st = Stream()
        while t_start < end:
            s3_file_path = PATH_FORMAT.format(year=t_start.year,
                                              julday=t_start.julday,
                                              station=site,
                                              network=net,
                                              location=loc,
                                              channel=comp)
            data = io.BytesIO()
            if self.debug:
                logger.debug('Requesting {}'.format(s3_file_path))
            _ = self.client.download_fileobj(Bucket=self.s3_bucket,
                                             Key=s3_file_path,
                                             Fileobj=data)
            data.seek(0)
            _st = read(data, dtype='float64')
            data.close()
            _st.trim(t_start, t_end, nearest_sample=False)
            st += _st
            t_start = t_end
            t_end = t_start + timedelta(days=1)
            t_end = min(t_end, end)

        st.merge(fill_value=self.fill_value)
        inv = get_instrument_response(net, site, loc,
                                      staxml_dir=self.staxml_dir,
                                      fdsn_urls=self.fdsn_urls)
        inv_dt = inv.select(location=loc, channel=comp,
                            starttime=t_start,
                            endtime=t_end)

        # Post-process to remove sensitivity and account for gaps
        pp = PostProcess(st=st, inv=inv, inv_dt=inv_dt,
                         startdate=start, enddate=end,
                         loc=loc, comp=comp, fill_value=self.fill_value)
        tr = pp.run_post_processing()
        return tr


def tilde_request(domain: str, name: str, station: str,
                  method: str, sensor: str, aspect: str,
                  startdate: datetime, enddate: datetime) -> pd.DataFrame:
    """
    Request data from the tilde API (https://tilde.geonet.org.nz/v3/api-docs/).
    See the tilde discovery tool for more information:
    https://tilde.geonet.org.nz/ui/data-discovery/

    Parameters
    ----------
    domain : str
        The domain of the data (e.g. 'manualcollect')
    name : str
        The name of the data (e.g. 'plume-SO2-gasflux')
    station : str
        The station code (e.g. 'WI000')
    method : str
        The method of the data (e.g. 'contouring')
    sensor : str
        The sensor of the data (e.g. 'MC01')
    aspect : str
        The aspect of the data (e.g. 'nil')
    startdate : date
        The start date of the data
    enddate : date
        The end date of the data

    Returns
    -------
    pd.DataFrame
        A pandas dataframe with the requested data
    """
    tilde_url = "https://tilde.geonet.org.nz/v3/data"
    # split the request into historic and latest data
    get_historic = True
    get_latest = False 
    startdate = startdate.astimezone(timezone.utc)
    enddate = enddate.astimezone(timezone.utc)
    _tstart = str(startdate.date())
    _tend = str(enddate.date())
    _today = datetime.now(timezone.utc).date()
    if enddate.date() > (_today - timedelta(days=29)):
        _tend = str((enddate.date() - timedelta(days=29)))
        get_latest = True 
    if startdate.date() > (_today - timedelta(days=29)):
        get_historic = False

    assert get_historic or get_latest, "Check start and end dates."
    
    if get_latest:
        latest = f"{tilde_url}/{domain}/{station}/{name}/{sensor}/{method}/{aspect}/latest/30d"
        rval = requests.get(latest, headers={'Accept':'text/csv'})
        if rval.status_code != 200:
            msg = f"Download of {name} for {station} failed with status code {rval.status_code}"
            msg += f" and url {latest}"
            raise ValueError(msg)
        df_latest = pd.read_csv(io.StringIO(rval.text), index_col='timestamp', parse_dates=['timestamp'],
                                usecols=['timestamp', 'value', 'error'],
                                date_format='ISO8601')
    if get_historic:
        historic = f"{tilde_url}/{domain}/{station}/{name}/{sensor}/{method}/{aspect}/"
        historic += f"{_tstart}/{_tend}"
        rval = requests.get(historic, headers={'Accept':'text/csv'})
        if rval.status_code != 200:
            msg = f"Download of {name} for {station} failed with status code {rval.status_code}"
            msg += f" and url {historic}"
            raise ValueError(msg)
        data = io.StringIO(rval.text)
        df_historic = pd.read_csv(data, index_col='timestamp', parse_dates=['timestamp'],
                                usecols=['timestamp', 'value', 'error'],
                                date_format='ISO8601')
        if get_latest and len(df_latest) > 0:
            df = df_historic.combine_first(df_latest)
        else:
            df = df_historic
    else:
        df = df_latest
    df.rename(columns={'value': 'obs', 'error': 'err'}, inplace=True)
    df.index.name = 'dt'
    return df.loc[startdate:enddate]


class GeoMagWaveforms(WaveformBaseclass):

    def __init__(self, base_url, method: str, aspect: str,
                 name: str):
        self.base_url = base_url
        self.method = method
        self.aspect = aspect
        self.name = name
        self.domain = 'geomag'

    def get_waveforms(self, net: str, site: str, loc: str, comp: str,
                     startdate: date, enddate: date) -> pd.DataFrame:
        df_ = tilde_request(domain=self.domain, name=self.name, station=site,
                           method=self.method, sensor=loc, aspect=self.aspect,
                           startdate=startdate.datetime,
                           enddate=enddate.datetime)

        df = df_.reindex(pd.date_range(start=df_.index[0], end=df_.index[-1], freq='60s'))
        df.fillna(method='ffill', inplace=True)
        stats = {'network': net, 'station': site, 'location': loc, 'channel': comp,
                 'npts':len(df), 'sampling_rate':1/60}
        stats['starttime'] = UTCDateTime(df.index[0])
        tr = Trace(data=df['obs'].values, header=stats)
        return tr 


class DataSource:

    def __init__(self, clients, chunk_size=None, cache_dir=None):
        self.chunk_size = chunk_size
        
        if len(clients) < 1:
            msg = "At least one data client is required."
            raise ValueError(msg)

        # Test that all waveform clients are derived from WaveformBaseclass
        for client in clients:
            if not issubclass(client.__class__, WaveformBaseclass):
                msg = "All clients must be derived from WaveformBaseclass."
                raise ValueError(msg)

        self.clients = clients
        self.cache_dir = os.path.join(os.environ['HOME'], 'zizou_cache')
        if cache_dir is not None:
            self.cache_dir = cache_dir

        os.makedirs(self.cache_dir, exist_ok=True)
        self.cache_client = SDS_Client(self.cache_dir)

    def to_sds(self, tr: Trace):
        """
        Save trace to SDS directory.
        """
        start = tr.stats.starttime
        end = tr.stats.endtime
        sds_fmtstr = os.path.join(
            "{year}", "{network}", "{station}", "{channel}.{sds_type}",
            "{network}.{station}.{location}.{channel}.{sds_type}.{year}.{doy:03d}")

        current_date = start.date
        while current_date <= end.date:
            _tr = tr.slice(UTCDateTime(current_date), UTCDateTime(current_date) + 86400)
            # print(_tr.stats.starttime, _tr.stats.endtime)
            fullpath = sds_fmtstr.format(year=_tr.stats.starttime.year,
                                        doy=_tr.stats.starttime.julday,
                                        sds_type="D", **_tr.stats)
            fullpath = os.path.join(self.cache_dir, fullpath)
            dirname, _ = os.path.split(fullpath)
            os.makedirs(dirname, exist_ok=True)
            logger.debug(f"writing {_tr} to {fullpath}")
            _tr.write(fullpath, format="MSEED")
            current_date += timedelta(days=1)

    def get_waveforms(self, net, site, loc, comp, start, end, cache=False):
        t_start = start
        t_end = end 
        if self.chunk_size is not None:
            t_end = t_start + self.chunk_size
            t_end = min(t_end, end)
        while t_end <= end:
            tr = Trace()
            try:
                # First check the cache
                st = self.cache_client.get_waveforms(net, site, loc, comp, t_start,
                                                    t_end, dtype='float64')
                tr = st.merge(fill_value=np.nan)[0]
                t_diff = int(t_end - t_start)
                if abs(t_diff - (tr.stats.npts - 1)*tr.stats.delta) > 1:
                    raise IndexError
                tr.stats['cached'] = True
            except (IndexError, AttributeError) as e:
                msg = "Data for {} between {} and {} not found in cache."
                logger.debug(msg.format('.'.join((net, site, loc, comp)),
                                        t_start, t_end))
                for client in self.clients:
                    try:
                        tr = client.get_waveforms(net, site, loc, comp, t_start, t_end)
                    except Exception as e:
                        logger.info(e)
                        continue
                    else:
                        break
                if not tr:
                    msg = "No data found for {}"
                    logger.error(msg.format('.'.join((net, site, loc, comp))))
                else:
                    if cache:
                        self.to_sds(tr)
            yield tr

            if self.chunk_size is not None:
                t_start = t_end 
                cond1 = (end - t_end) < self.chunk_size
                cond2 = end > t_end 
                if cond1 and cond2:
                    last_chunk = end - t_end
                    t_end = t_start + last_chunk
                else:
                    t_end = t_start + self.chunk_size
            else:
                # go beyond the end to stop the loop
                t_end = end + 1.


class VolcanoMetadata:
    """
    Read volcano/station/channel metadata
    """

    def __init__(self, configfile):
        try:
            with open(configfile, 'r') as fh:
                cfg = yaml.safe_load(fh)
        except OSError:
            cfg = yaml.safe_load(configfile)
        self.data = cfg['metadata']

    def get_available_volcanoes(self):
        return [vol['name'] for vol in self.data['volcano']]

    def get_available_streams(self, volcano):
        """
        Get available sensor streams for a given volcano.

        Parameters
        ----------
        volcano : str
            Volcano name 

        Returns
        -------
        list 
            List of streams in the format 'NET.STA.LOC.CHAN' 
        """
        _streams = [] 
        for _v in self.data['volcano']:
            if _v['name'].lower() == volcano.lower():
                for _n in _v['network']:
                    net_code = _n['net_code']
                    for _s in _n['stations']:
                        sta_code = _s['sta_code']
                        location = _s['location']
                        for _c in _s['channel']:
                            chan_code = _c['code']
                            _stream = '.'.join((net_code, sta_code, location, chan_code))
                            _streams.append(_stream)
        return _streams

    def get_site_information(self, sitename):
        """
        Get site information for a given site. Note 
        that if the same site name exists at multiple
        volcanoes, the first match will be returned.

        Parameters
        ----------
        sitename : str 
            Three-letter site code. 

        Returns
        -------
        dict 
            Dictionary with site information as follows:
            {'volcano': str, 'site': str, 'channels': list, 
             'latitude': float, 'longitude': float, 'starttime': datetime}            
        """
        for _v in self.data['volcano']:
            for _n in _v['network']:
                for _s in _n['stations']:
                    if _s['sta_code'] == sitename:
                        try:
                            starttime = datetime.strptime(_s['starttime'],
                                                        "%Y-%m-%dT%H:%M:%SZ")
                        except KeyError:
                            starttime = None

                        try:
                            channels = [c['code'] for c in _s['channel']]
                        except KeyError:
                            channels = []
                        return dict(volcano=_v['name'],
                                    site=_s['sta_code'],
                                    channels=channels,
                                    latitude=float(_s['latitude']),
                                    longitude=float(_s['longitude']),
                                    starttime=starttime)
        return None

    def get_eruption_dates(self, volcano):
        """
        Get eruption dates for a given volcano.

        Parameters
        ----------
        volcano : str 
            Volcano name. 

        Returns
        -------
        list 
            List of datetime.date objects representing eruption dates. 
        """
        eruption_dates = []
        metadata = [v for v in self.data['volcano']
                    if v['name'] == volcano]
        er = metadata[0]["eruptions"]
        if er:
            for e in er:
                d, m, y = e.split('-')
                eruption_dates.append(date(int(y), int(m), int(d)))
        return eruption_dates

    def get_unrest_periods(self, volcano):
        """
        Get unrest periods for a given volcano.

        Parameters
        ----------
        volcano : str 
            Volcano name. 

        Returns
        -------
        list 
            List of datetime.date objects representing start and end dates. 
        """
        unrest_dates = []
        metadata = [v for v in self.data['volcano']
                    if v['name'] == volcano]
        ur = metadata[0]["unrest periods"]
        if ur:
            for u in ur:
                ds, ms, ys = u['starttime'].split('-')
                de, me, ye = u['endtime'].split('-')
                unrest_dates.append([date(int(ys), int(ms), int(ds)),
                                     date(int(ye), int(me), int(de))])
        return unrest_dates


if __name__ == '__main__':
    import doctest
    doctest.testmod()
